{{- $showPageLastmod := (or (default false .Page.Params.geekdocPageLastmod) (default false .Site.Params.geekdocPageLastmod)) -}}

{{- if $showPageLastmod -}}
<span class="flex align-center no-wrap">
    <svg class="gdoc-icon gdoc_date"><use xlink:href="#gdoc_date"></use></svg>
    <time datetime="{{ .Lastmod.Format "2006-01-02T15:04:05Z07:00" | safeHTML }}">
        {{ if .Lastmod.After (.Date.AddDate 0 0 1) }}
        {{ i18n "posts_update_prefix" }}
        {{ end }}
        {{ .Lastmod.Format "Jan 2, 2006" }}
    </time>
</span>
{{- end -}}
